#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <immintrin.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/if_packet.h>
#include <math.h>
#include <netdb.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "private/trx_driver.h"

#define DEBUG
#define SSE4 /* define if CPU supports SSE4.1 */

#include "utils.c"
#include "ring_buffer.c"

// Update period for packets per second statistic counter
#define PPS_UPDATE_PERIOD INT64_C(1000000000)
#define STAT_INT_LEN "9"

#define N_SAMPLES (32)
#define MIN_PACKET_SIZE 64
#define MAX_CHANNELS 4
#define ETHERNET_HEADER 14
#define ECPRI_COMMON_HEADER 4
#define ECPRI_IQ_HEADER (ECPRI_COMMON_HEADER + 4)
#define ORAN_HEADER 8
#define PACKET_HEADER (ETHERNET_HEADER + ECPRI_IQ_HEADER + ORAN_HEADER)
#define IQ_PAYLOAD 8832
#define PACKET_SIZE (PACKET_HEADER + IQ_PAYLOAD)

/* eCPRI frame structure

Ethernet header: 14 bytes
     6         6          2
+---------+---------+-----------+
| MAC dst | MAC src | Ethertype |
+---------+---------+-----------+

eCPRI common header: 4 bytes

     4b        3b        1b        1         2     
+---------+----------+--------+---------+---------+
| Version | Reserved | Concat | Message | Payload |
|         |          |        |  Type   |  Size   |
+---------+----------+--------+---------+---------+

IQ data transfer message header: 4 bytes

    2        2          
+-------+--------+------------+
| PC_ID | SEQ_ID | IQ Samples |
+-------+--------+------------+

eCPRI concatenated frames

      4                             3          4
+-----------+------------------+---------+-----------+------------------+
|   eCPRI   |                  |         |   eCPRI   |                  |
|  Common   | eCPRI Payload #0 | Padding |  Common   | eCPRI Payload #1 |
| Header #0 |                  |         | Header #1 |                  |
+-----------+------------------+---------+-----------+------------------+

IQ sample pair: 32 bit (4 bytes)
IQ / second / antenna: 30 720 000
IQ payload per frame: 8832 bytes = 2208 IQ samples
IQ throughput (4 * 20 MHz): 3 932 160 000 bytes

1 radio frame = 10 subframes
1 subframe = 2 slots
1 slot = 7 symbols

One ethernet frame = 8862

        14                   4                  4               8          8832
+-----------------+---------------------+----------------+-------------+------------+
| Ethernet header | eCPRI common header | IQ data header | oRAN header | IQ payload |
+-----------------+---------------------+----------------+-------------+------------+

*/

typedef struct {
  float re;
  float im;
} Complex;

typedef struct {
  const uint8_t * rrh_mac;
  const uint8_t * bbu_mac;
  const uint8_t * bbu_if;
  const char * log_directory;
  int recv_affinity;
  int send_affinity;
  int encode_affinity;
  int decode_affinity;
  int statistic_affinity;

  int ecpri_period;
  int flow_id;
  int frame_frequency;

  int trx_buf_size;
  int txrx_buf_size;

  int trace_rx;
  int trace_tx;
  int trace_offset;

  int monitor_pps;
  int monitor_trigger_duration;

  int start_sending;
  int start_receiving;
  int rx_n_channel;
  int tx_n_channel;
  int statistics_refresh_rate_ns;
  int sample_rate;
} TRXEcpriState;

typedef struct {
  int64_t counter;
  int64_t pps_counter;
  int64_t pps_ts;
  int64_t pps;
} counter_stat_t;

/* Proprietary code:
     - compression / decompression of IQ samples
     - fast conversion between int16_t and float
*/
//#include "private/bf1_avx2.c"
#include "private/bf1_sse4.c"

// Buffers
static ring_buffer_t rx_rbuf; // Received packets, ethernet frames
static ring_buffer_t trxr_rbuf[MAX_CHANNELS]; // Decoded IQ samples
static ring_buffer_t tx_rbuf; // Packets to send, ethernet frames
static ring_buffer_t trxw_rbuf[MAX_CHANNELS]; // Uncompressed IQ samples

// Counters
static volatile counter_stat_t recv_counter; // frames received from RRH
static volatile counter_stat_t decode_counter; // decoded frames
static volatile counter_stat_t read_counter; // frames passed to amarisoft stack
static volatile counter_stat_t write_counter; // samples to write from TRX
static volatile counter_stat_t encode_counter; // encoded frames
static volatile counter_stat_t sent_counter; // frames sent to RRH
static volatile counter_stat_t rx_drop_counter; // frames sent to RRH
static volatile counter_stat_t tx_drop_counter; // frames sent to RRH

// Network
static volatile uint8_t seq_id;
static volatile uint8_t frame_id;
static volatile uint8_t sub_frame_id;
static volatile uint8_t slot_id;
static volatile uint8_t symbol_id;
static int send_sockfd;
static int recv_sockfd;
static struct sockaddr_ll connect_sk_addr;
static uint8_t ecpri_iq_header[ECPRI_IQ_HEADER + ORAN_HEADER];
static uint8_t packet_header[PACKET_HEADER]; // ethernet + ecpri + iq header

static void print_stats(FILE * f, int print_header) {
  if(print_header) {
    fprintf(f, 
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s    "
        "%" STAT_INT_LEN "s    "
        "%" STAT_INT_LEN "s    "
        "%" STAT_INT_LEN "s    "
        "%" STAT_INT_LEN "s    "
        "%" STAT_INT_LEN "s    "
        "\n",
        "rx dropped",
        "tx dropped",
        "received",
        "decode",
        "read",
        "write",
        "encode",
        "sent",
        "received pps",
        "decode pps",
        "read pps",
        "write pps",
        "encode pps",
        "sent pps");
  }
  fprintf(f, 
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 " "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "%" STAT_INT_LEN "" PRIi64 "pps "
      "\n",
      rx_drop_counter.counter,
      tx_drop_counter.counter,
      recv_counter.counter,
      decode_counter.counter,
      read_counter.counter,
      write_counter.counter,
      encode_counter.counter,
      sent_counter.counter,
      recv_counter.pps,
      decode_counter.pps,
      read_counter.pps,
      write_counter.pps,
      encode_counter.pps,
      sent_counter.pps);
}

static void log_exit(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " EXIT [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  fprintf(stderr, "%s\n", line);

  // Dump useful information
  print_stats(stderr, 1);
  fprintf(stderr, "TX RBUF: ri %d wi %d ra %d wa %d\n",
    tx_rbuf.read_index, tx_rbuf.write_index, rbuf_read_amount(&tx_rbuf),
    rbuf_write_amount(&tx_rbuf));
  fprintf(stderr, "RX RBUF: ri %d wi %d ra %d wa %d\n",
    rx_rbuf.read_index, rx_rbuf.write_index, rbuf_read_amount(&rx_rbuf),
    rbuf_write_amount(&rx_rbuf));
  fprintf(stderr, "TRXW RBUF: ri %d wi %d ra %d wa %d\n",
    trxw_rbuf[0].read_index, trxw_rbuf[0].write_index, rbuf_read_amount(&trxw_rbuf[0]),
    rbuf_write_amount(&trxw_rbuf[0]));
  fprintf(stderr, "TRXR RBUF: ri %d wi %d ra %d wa %d\n",
    trxr_rbuf[0].read_index, trxr_rbuf[0].write_index, rbuf_read_amount(&trxr_rbuf[0]),
    rbuf_write_amount(&trxr_rbuf[0]));

  fflush(stdout);
  fflush(stderr);
  exit(EXIT_FAILURE);
}

static void init_counter(volatile counter_stat_t * c) {
  c->counter = 0;
  c->pps_counter = 0;
  c->pps_ts = 0;
  c->pps = 0;
}
static void update_counter_pps(volatile counter_stat_t * c) {
  struct timespec _ts;
  int64_t ts;
  clock_gettime(CLOCK_TAI, &_ts);
  ts = ts_to_int(_ts);
  if((ts - c->pps_ts) > PPS_UPDATE_PERIOD) {
    if(c->pps_ts)
      c->pps = ((c->counter - c->pps_counter) * NSEC_PER_SEC) / (ts - c->pps_ts);
    c->pps_counter = c->counter;
    c->pps_ts = ts;
  }
}
static void update_counter(volatile counter_stat_t * c, int64_t v) {
  c->counter += v;
}

static void *recv_thread(void *p) {

  cpu_set_t mask;
  TRXEcpriState * s = (TRXEcpriState *) p;
  int ret;

  log_info("RECV_THREAD", "Thread init");
  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->recv_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->recv_affinity);

  for(;;) {

    struct mmsghdr msgh[4000];
    struct iovec msgv[4000];

    memset(msgv, 0, sizeof(msgv));
    memset(msgh, 0, sizeof(msgh));

    for(int j = 0; j < ecpri_period_mult; j++) {
      msgv[j].iov_base = RBUF_WRITE(rx_rbuf, uint8_t);
      msgv[j].iov_len = rx_rbuf.len;
      msgh[j].msg_hdr.msg_iov = &msgv[j];
      msgh[j].msg_hdr.msg_iovlen = 1;
      rbuf_update_write_index(&rx_rbuf);
    }

    ret = recvmmsg(recv_sockfd, msgh, ecpri_period_mult, 0, NULL);
    if(ret == -1)
      error(EXIT_FAILURE, errno, "recvmmsg error");
    if(ret != ecpri_period_mult)
      log_error("RECV_THREAD", "recvmmsg received %d messages instead of %d\n", ret, ecpri_period_mult);
  }
  pthread_exit(EXIT_SUCCESS);
}

static void *send_thread(void *p) {

	cpu_set_t mask;
  struct timespec initial, next;
  struct timespec t1[4000];
  struct timespec t2[4000];
  int k = 0;
  TRXEcpriState * s = (TRXEcpriState *) p;
  struct mmsghdr msgh[4000];
  struct iovec msgv[4000];
  log_info("SEND_THREAD", "Thread init");

	// Set thread CPU affinity
	CPU_ZERO(&mask);
	CPU_SET(s->send_affinity, &mask);
	if (sched_setaffinity(0, sizeof(mask), &mask))
		error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->send_affinity);

  memset(msgv, 0, sizeof(msgv));
  memset(msgh, 0, sizeof(msgh));

  for(int j = 0; j < ecpri_period_mult; j++) {
    msgh[j].msg_hdr.msg_name = &connect_sk_addr;
    msgh[j].msg_hdr.msg_namelen = sizeof(connect_sk_addr);
    msgh[j].msg_hdr.msg_iov = &msgv[j];
    msgh[j].msg_hdr.msg_iovlen = 1;
  }

  pthread_mutex_lock(&tx_ready_mutex);
  pthread_cond_wait(&tx_ready_cond, &tx_ready_mutex);
  pthread_mutex_unlock(&tx_ready_mutex);

  clock_gettime(CLOCK_TAI, &initial);
  for(int64_t i = 1;; i++) {
    int ret, msg_sent;

#ifdef DEBUG
    if(i > SEND_LIMIT) {
      int64_t d, dt;
      clock_gettime(CLOCK_TAI, &next);
      d = calcdiff_ns(next, initial);
      for(int j = 0; j < k; j++) {
        dt = calcdiff_ns(t2[j], t1[j]);
        log_debug("SEND_THREAD", "%" PRIi64, dt);
      }
      log_debug("SEND_THREAD", "Packets sent: %" PRIi64, sent_frame_count);
      log_debug("SEND_THREAD", "Duration: %" PRIi64, d);
      log_debug("SEND_THREAD", "ecpri_period_mult: %" PRIi64, ecpri_period_mult);
      log_debug("SEND_THREAD", "FRAME_FREQ: %" PRIi64, FRAME_FREQ);
      exit(EXIT_SUCCESS);
    }
#endif

    next = initial;
    // Multiply by i everytime to prevent any frequence drift
    add_ns(&next, (ecpri_period_mult * NSEC_PER_SEC * i) / FRAME_FREQ);

    for(int j = 0; j < ecpri_period_mult; j++) {
      msgv[j].iov_base = RBUF_READ(tx_rbuf, uint8_t);
      msgv[j].iov_len = tx_rbuf.len;
      rbuf_update_read_index(&tx_rbuf);
    }

    for(msg_sent = 0; msg_sent < ecpri_period_mult;) {
#ifdef DEBUG
      clock_gettime(CLOCK_TAI, &t1[k]);
#endif
      ret = sendmmsg(send_sockfd, msgh + msg_sent, (ecpri_period_mult - msg_sent), 0);
#ifdef DEBUG
      clock_gettime(CLOCK_TAI, &t2[k++]);
#endif
      if(ret <= 0)
        error(EXIT_FAILURE, errno, "sendmmsg error (returned %d)", ret);
      msg_sent += ret;
      sent_frame_count += ret;
    }

    pthread_mutex_lock(&tx_mutex);
    pthread_cond_signal(&tx_cond);
    pthread_mutex_unlock(&tx_mutex);

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
  pthread_exit(EXIT_SUCCESS);
}

static void *encode_thread(void *p) {

  cpu_set_t mask;
  TRXEcpriState * s = (TRXEcpriState *) p;
  int64_t target_counter = 0;
  struct timespec next;
  int reset_encode_counter = 1;
  uint8_t * data;

  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->encode_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->encode_affinity);

  for(int64_t i = 0;; i++) {
    int64_t to_write, to_read;
    uint16_t * 16_bit;
    int j;

    // If we have frames to encode (is there space in TX buffer)
    // And if there are frames from trx_write callback to encode
    to_write = rbuf_write_amount(&tx_rbuf);
    to_read = rbuf_read_amount(&trxw_rbuf[0]);
    if(to_write && to_read) {

      for(uint16_t antenna_id = 0 ; antenna_id < s->tx_n_channel; antenna_id++) {
        data = rbuf_write(&tx_rbuf);

        j = ETHERNET_HEADER + ECPRI_COMMON_HEADER;
        // PC_ID
        *((uint16_t *) data[j + 0]) = htons(antenna_id);
        // SEQ_ID
        *((uint16_t *) data[j + 2]) = htons(seq_id++ << 8 + 0x80);
        j = ETHERNET_HEADER + ECPRI_IQ_HEADER;
        // ORAN counters
        data[j + 1] = frame_id;
        16_bit = (uint16_t *) data[j + 2];
        *16_bit  = sub_frame_id << 8;
        *16_bit |= slot_id << 4 ;
        *16_bit |= symbol_id;

        // 8832 bytes of IQ samples

        // Add IQ_PAYLOAD to the current packet
        int write_count;
        int count_left = IQ_PAYLOAD;
        while((write_count = rbuf_contiguous_copy(&trxw_rbuf[antenna_id], &tx_rbuf,
                                                  count_left))) {
          memcpy(data[PACKET_HEADER] + (IQ_PAYLOAD - count_left),
                 rbuf_read(&trxw_rbuf[antenna_id], IQ_PAYLOAD - count_left),
                 write_count);

          rbuf_increment_read(&trxw_rbuf[antenna_id], write_count);

          tx_rbuf.write_index = (tx_rbuf.write_index + write_count) % tx_rbuf.buf_len;
          trxw_rbuf[0].read_index = (trxw_rbuf[0].read_index + write_count) % \
                                    trxw_rbuf[0].buf_len;
          count_left -= write_count;
        }
        if(count_left)
          exit(EXIT_FAILURE);
      }
      // Increment counter once we wrote packets for all channels
      update_counter(&encode_counter, 1);

      // Update ORAN counters
      slot_id = (slot_id + 1) % 2;
      if(!slot_id) {
        sub_frame_id = (sub_frame_id + 1) % 10;
        if(!sub_frame_id) {
          frame_id = (frame_id + 1) % 100;
        }
      }
    }
  }
  pthread_exit(EXIT_SUCCESS);
}

static void *decompress_thread(void *p) {

	cpu_set_t mask;
  TRXEcpriState * s = (TRXEcpriState *) p;
  int rx_ready = 0;
  const float mult = 1. / 32767.;
  log_info("DECOMPRESS_THREAD", "Thread init");

	// Set thread CPU affinity
	CPU_ZERO(&mask);
	CPU_SET(s->decompress_affinity, &mask);
	if (sched_setaffinity(0, sizeof(mask), &mask))
		error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->decompress_affinity);

  for(;;) {
    int n = rbuf_read_amount(&rx_rbuf);
    if(n) {
      for(int j = 0; j < n; j++) {
        int16_t samples_int[256];
        const uint8_t * rx_samples = RBUF_READ(rx_rbuf, uint8_t) + 22;
        // TODO : analyze seq_id, ecpri packet type etc... ?
        // TODO : set rx_ready at some point (when ?)
        if(rx_ready) {
          memset(samples_int, 0, 512);

          decode_bf1(samples_int      , rx_samples      , 16);
          decode_bf1(samples_int + 64 , rx_samples  + 60, 16);
          decode_bf1(samples_int + 128, rx_samples + 120, 16);
          decode_bf1(samples_int + 192, rx_samples + 180, 16);
          int16_to_float(RBUF_WRITE(trx_read_rbuf, float), samples_int, 256, mult);

          rbuf_update_read_index(&rx_rbuf);
          rbuf_update_write_index(&trx_read_rbuf);
          sem_post(&trx_read_sem);
        }
      }
    }
    else {
      pthread_mutex_lock(&rx_mutex);
      pthread_cond_wait(&rx_cond, &rx_mutex);
      pthread_mutex_unlock(&rx_mutex);
    }
  }
  pthread_exit(EXIT_SUCCESS);
}

static void *statistic_thread(void *p) {
  struct timespec next, initial;
  int64_t recv_stop = 0;
  cpu_set_t mask;
  TRXEcpriState * s = (TRXEcpriState *) p;
  FILE * stats_file_desc;

  log_info("STATISTIC_THREAD", "Thread init");

  char stats_file_name[256];
  memset(stats_file_name, '\0', 256);
  sprintf(stats_file_name, "%s/ecpri.stats", s->log_directory);
  stats_file_desc = fopen(stats_file_name, "w+");

  if(!stats_file_desc)
    error(EXIT_FAILURE, errno, "Couldn't open %s\n", stats_file_name);

  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->statistic_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->statistic_affinity);

  clock_gettime(CLOCK_TAI, &initial);
  next = initial;
  for(int64_t i = 0;; i++) {
    add_ns(&next, s->statistics_refresh_rate_ns);
    if(s->trace_rx || s->trace_tx)
      trace_handler(initial, s);
    print_stats(stats_file_desc, (i % 50) == 0);
#ifdef DEBUG
    fprintf(stats_file_desc, 
        "%d %d %d %d %d %d %d %d\n", 
        rx_rbuf.write_index,
        rx_rbuf.read_index,
        trxr_rbuf[0].write_index,
        trxr_rbuf[0].read_index,
        trxw_rbuf[0].write_index,
        trxw_rbuf[0].read_index,
        tx_rbuf.write_index,
        tx_rbuf.read_index);
    fprintf(stats_file_desc, "TRXW RBUF: ri %d wi %d ra %d wa %d\n", trxw_rbuf[0].read_index, trxw_rbuf[0].write_index, rbuf_read_amount(&trxw_rbuf[0]), rbuf_write_amount(&trxw_rbuf[0]));
#endif
    fflush(stats_file_desc);

    update_counter_pps(&rx_drop_counter);
    update_counter_pps(&tx_drop_counter);
    update_counter_pps(&recv_counter);
    update_counter_pps(&decode_counter);
    update_counter_pps(&read_counter);
    update_counter_pps(&write_counter);
    update_counter_pps(&encode_counter);
    update_counter_pps(&sent_counter);
    if(s->monitor_pps) {
      if(recv_counter.pps > recv_pps_threshold) {
        recv_pps_threshold_hit = 1;
      }
      if(recv_pps_threshold_hit && recv_counter.pps < recv_pps_threshold) {
        struct timespec _ts;
        int64_t ts;
        clock_gettime(CLOCK_MONOTONIC, &_ts);
        ts = ts_to_int(_ts);
        if((recv_stop && ((ts - recv_stop) > (s->monitor_trigger_duration) * INT64_C(1000000000)))) {
          if(s->monitor_pps == 1)
            log_exit("MONITOR", "Stopped recieving packets, restarting...");
          log_info("MONITOR", "Stopped recieving packets, sending again...");
          sync_complete = 0;
          recv_stop = 0;
        }
        if(!recv_stop)
          recv_stop = ts;
      }
    }

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
  pthread_exit(EXIT_SUCCESS);
}

static int start_threads(TRXEcpriState * s) {
  pthread_t recv_pthread;
  pthread_t send_pthread;
  pthread_t encode_pthread;
  pthread_t decode_pthread;
  pthread_t statistic_pthread;
  struct sched_param recv_param;
  struct sched_param send_param;
  struct sched_param encode_param;
  struct sched_param decode_param;
  struct sched_param statistic_param;
  pthread_attr_t recv_attr;
  pthread_attr_t send_attr;
  pthread_attr_t encode_attr;
  pthread_attr_t decode_attr;
  pthread_attr_t statistic_attr;

  log_info("TRX_ECPRI", "Starting threads");

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&recv_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  // Set a specific stack size
  if (pthread_attr_setstacksize(&recv_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&recv_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  recv_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&recv_attr, &recv_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&recv_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&send_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&send_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&send_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  send_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&send_attr, &send_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&send_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&encode_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&encode_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&encode_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  encode_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&encode_attr, &encode_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&encode_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&decode_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&decode_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&decode_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  decode_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&decode_attr, &decode_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&decode_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&statistic_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&statistic_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&statistic_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  statistic_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&statistic_attr, &statistic_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&statistic_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_create(&statistic_pthread, NULL, statistic_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create statistic thread");
  usleep(1000 * 20);
  if (pthread_create(&encode_pthread, NULL, encode_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create encode thread");
  usleep(1000 * 20);
  if (pthread_create(&decode_pthread, NULL, decode_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create decode thread");
  usleep(1000 * 20);
  if (pthread_create(&send_pthread, NULL, send_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create send thread");
  usleep(1000 * 500);
  if (pthread_create(&recv_pthread, NULL, recv_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create recv thread");

  return 0;
}

int start(TRXEcpriState * s) {
  uint8_t dst_mac[6];
  uint8_t src_mac[6];
  struct ether_header *eh = (struct ether_header *) packet_header;
  int if_index;

  log_debug("TRX_ECPRI", "raw socket setup");

	//set_latency_target();

  seq_id = 0;
  init_counter(&rx_drop_counter);
  init_counter(&tx_drop_counter);
  init_counter(&recv_counter);
  init_counter(&decode_counter);
  init_counter(&read_counter);
  init_counter(&write_counter);
  init_counter(&encode_counter);
  init_counter(&sent_counter);

  init_rbuf(&rx_rbuf, "RX ring buffer", s->txrx_buf_size, PACKET_SIZE);
  init_rbuf(&tx_rbuf, "TX ring buffer", s->txrx_buf_size, PACKET_SIZE);

  for(int i = 0; i < s->tx_n_channel; i++) {
    char name[256];
    sprintf(name, "TRXWrite Ring Buffer %d", i);
    init_rbuf(trxw_rbuf[i], name, s->trx_buf_size, sizeof(Complex) * 32);
  }
  for(int i = 0; i < s->rx_n_channel; i++) {
    char name[256];
    sprintf(name, "TRXRead Ring Buffer %d", i);
    init_rbuf(trxr_rbuf[i], name, s->trx_buf_size, sizeof(Complex) * 32);
  }

  memset((uint8_t *) packet_header,   0, PACKET_HEADER);
  memset((uint8_t *) ecpri_iq_header, 0, ECPRI_IQ_HEADER + ORAN_HEADER);

  if(sscanf((char *) s->rrh_mac, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx%*c",
        &d_addr.addr_bytes[0],
        &d_addr.addr_bytes[1],
        &d_addr.addr_bytes[2],
        &d_addr.addr_bytes[3],
        &d_addr.addr_bytes[4],
        &d_addr.addr_bytes[5]) != 6)
    fprintf(stderr, "Invalid RRH MAC address\n");
  if(sscanf((char *) s->bbu_mac, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx%*c",
        &s_addr.addr_bytes[0],
        &s_addr.addr_bytes[1],
        &s_addr.addr_bytes[2],
        &s_addr.addr_bytes[3],
        &s_addr.addr_bytes[4],
        &s_addr.addr_bytes[5]) != 6)
    fprintf(stderr, "Invalid BBU MAC address\n");

  if (!(if_index = if_nametoindex(s->bbu_if))) {
    perror("if_nametoindex");
    return 1;
  }

  if ((send_sockfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) == -1) {
      perror("Socket Error");
      return 1;
  }
  if ((recv_sockfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) == -1) {
      perror("Socket Error");
      return 1;
  }

  connect_sk_addr.sll_ifindex = if_index;
  connect_sk_addr.sll_halen = ETH_ALEN;
  for(int i = 0; i < 6; i++)
    connect_sk_addr.sll_addr[i] = dst_mac[i];

  for(int i = 0; i < 6; i++)
    eh->ether_shost[i] = src_mac[i];
  for(int i = 0; i < 6; i++)
    eh->ether_dhost[i] = dst_mac[i];

  /* Ethertype field */
  eh->ether_type = htons(0xaefe);

  /* Common Header */
  ecpri_iq_header[0] = 0x10; // Version 0x1, Reserved = 0, C = 0
  ecpri_iq_header[1] = 0x00; // Message type (IQ data)
  /* IQ message header */
  *((uint16_t *) (ecpri_iq_header + ECPRI_COMMON_HEADER)) = htons(s->flow_id);

  /* ORAN HEADER */
  ecpri_iq_header[ECPRI_IQ_HEADER] = 0x90;
  ecpri_iq_header[ECPRI_IQ_HEADER + 3] = 0x00;
  ecpri_iq_header[ECPRI_IQ_HEADER + 4] = 0x00;
  ecpri_iq_header[ECPRI_IQ_HEADER + 5] = 0x00;
  ecpri_iq_header[ECPRI_IQ_HEADER + 6] = 0x00;

  memcpy(packet_header + ETHERNET_HEADER,
         ecpri_iq_header,
         ECPRI_IQ_HEADER + ORAN_HEADER);

  for(int i = 0; i < tx_rbuf.buf_len; i+= PACKET_SIZE)
    memcpy(((uint8_t *) tx_rbuf.buffer) + i, packet_header, PACKET_HEADER);

  start_threads(s);
  return 0;
}

static void trx_ecpri_end(TRXState *s1)
{
  log_info("TRX_ECPRI", "End");
  TRXEcpriState *s = s1->opaque;
  free(s);
}

static int64_t prev_timestamp = 0;
static int64_t prev_count = 0;

/*
   Callback for processing TRX samples from Amarisoft
   Write count samples from __samples into TRX write buffer, exit as soon as possible
 
   - count: how much IQ samples to write (multiple of 32)
   - timestamp: should be equal to the amount of written IQ samples
   - __samples: can be null if zeroes are to be written (TDD only)
*/
static void trx_ecpri_write(TRXState *s1, trx_timestamp_t timestamp, const void **__samples, int count, int tx_port_index, TRXWriteMetadata *md)
{
  int64_t count_left, ts, nc, offset;
  float ** _samples = (float **) __samples;
  TRXEcpriState *s = s1->opaque;

  log_debug("TRX_ECPRI_WRITE", "trx_ecpri_write, count = %ld", count / N_SAMPLES);
  
  // Consistency check
  if(prev_count && ((timestamp - prev_timestamp) != prev_count)) {
    log_exit("TRX_ECPRI_WRITE",
      "Gap between timestamps: prev_ts %li ts %li prev_count %li count %li diff_ts %li",
      prev_timestamp, timestamp, prev_count, count, (timestamp - prev_timestamp));
  }
  prev_timestamp = timestamp; prev_count = count;

  // Drop samples if we don't have enough space in trx write buffer
  if( (count * sizeof(Complex)) > rbuf_write_amount(&trxw_rbuf[0]) ) {
    //log_exit("TRX_ECPRI_WRITE",
    //  "Not enough space to write in trxw_rbuf (count = %d)", count);
    update_counter(&tx_drop_counter, count / N_SAMPLES);
    return;
  }

  offset = 0;
  count_left = count;
  while((nc = rbuf_contiguous_copy(NULL, &trxw_rbuf[0], count_left * sizeof(Complex)))) {
    for(int i = 0; i < s->tx_n_channel; i++) {
      if(__samples)
        memcpy(
          ((uint8_t *) trxw_rbuf[i].buffer) + trxw_rbuf[0].write_index,
          ((uint8_t *) _samples[i]) + offset,
          nc);
      else
        memset(
          ((uint8_t *) trxw_rbuf[i].buffer) + trxw_rbuf[0].write_index,
          0,
          nc);
    }
    rbuf_increment_write(&trxw_rbuf, nc, 0);
    count_left -= nc / sizeof(Complex);
    offset += nc;
  }
  update_counter(&write_counter, count / N_SAMPLES);
}

/*
   Callback for sending TRX samples to Amarisoft
   Put count IQ samples from TRX read buffer in __samples, return as soon as possible
 
   - count: how much IQ samples to read (multiple of 32)
   - ptimestamp: number of total read IQ samples
   - __samples: IQ samples returned
*/
static int trx_ecpri_read(TRXState *s1, trx_timestamp_t *ptimestamp, void **__samples,
                          int count, int rx_port_index, TRXReadMetadata *md)
{
  int64_t nc, count_left, offset;
  float ** _samples = (float **) __samples;
  TRXEcpriState *s = s1->opaque;

  log_debug("TRX_ECPRI_READ", "count = %ld (%li)",
    count / N_SAMPLES, read_counter.counter);

  // Wait to have enough sampels in TRX read buffer
  while( rbuf_read_amount(&trxr_rbuf[0]) < (count * sizeof(Complex)) );
  
  sync_complete = 1;

  offset = 0;
  count_left = count;
  while((nc = rbuf_contiguous_copy(&trxr_rbuf[0], NULL, count_left * sizeof(Complex)))) {
    for(int i = 0; i < s->rx_n_channel; i++ ) {
      memcpy(
        ((uint8_t*) _samples[i]) + offset,
        ((uint8_t *) trxr_rbuf[i].buffer) + trxr_rbuf[0].read_index,
        nc);
    }
    rbuf_increment_read(&trxr_rbuf, nc);
    count_left -= nc / sizeof(Complex);
    offset += nc;
  }
  *ptimestamp = (read_counter.counter) * N_SAMPLES;
  update_counter(&read_counter, count / N_SAMPLES);

  return count;
}

/* This function can be used to automatically set the sample
   rate. Here we don't implement it, so the user has to force a given
   sample rate with the "sample_rate" configuration option */
static int trx_ecpri_get_sample_rate(TRXState *s1, TRXFraction *psample_rate,
    int *psample_rate_num, int sample_rate_min)
{
  return -1;
}

static int trx_ecpri_start(TRXState *s1, const TRXDriverParams *params)
{
  TRXEcpriState *s = s1->opaque;

  log_info("TRX_ECPRI_START", "Start");

  log_info("TRX_ECPRI_START", "trx_api_version: %d", s1->trx_api_version);
  log_info("TRX_ECPRI_START", "config file: %s", s1->path);
  log_info("TEST-DPDK-ECPRI", "bbu-mac: %s, rrh-mac: %s, bbu-if: %s", s->bbu_mac, s->rrh_mac, s->bbu_if);

  s->sample_rate = params->sample_rate[0].num / params->sample_rate[0].den;

  start(s);

  return 0;
}

void dummy_enb_init(TRXState *s1, TRXEcpriState *s) {

  s1->trx_write_func2 = trx_ecpri_write;
  s1->trx_read_func2 = trx_ecpri_read;
  start(s);
}

/* Called to start the tranceiver. Return 0 if OK, < 0 if  */
int trx_start_func(TRXState *s, const TRXDriverParams *p) {
  log_info("DEBUG", "trx_start_func");
  return 0;
}

/* Deprecated, use trx_write_func2 instead.
   Write 'count' samples on each channel of the TX port
   'tx_port_index'. samples[0] is the array for the first
   channel. timestamp is the time (in samples) at which the first
   sample must be sent. When the TRX_WRITE_FLAG_PADDING flag is
   set, samples is set to NULL. It indicates that no data should
   be sent (TDD receive time). TRX_WRITE_FLAG_END_OF_BURST is set
   to indicate in advance that the next write call will have the
   TRX_WRITE_FLAG_PADDING flag set. Note:
   TRX_WRITE_FLAG_END_OF_BURST and TRX_WRITE_FLAG_PADDING are
   never set simultaneously.
 */
void trx_write_func(TRXState *s, trx_timestamp_t timestamp, const void **samples, int count, int flags, int tx_port_index) {
  log_info("DEBUG", "**");
}

/* Deprecated, use trx_read_func2 instead.
   Read 'count' samples from each channel. samples[0] is the array
   for the first channel. *ptimestamp is the time at which the
   first samples was received. Return the number of sample read
   (=count).

   Note: It is explicitely allowed that the application calls
   trx_write_func, trx_read_func, trx_set_tx_gain_func and
   trx_set_rx_gain_func from different threads.
*/
int trx_read_func(TRXState *s, trx_timestamp_t *ptimestamp, void **samples, int count, int rx_port_index) {
  log_info("DEBUG", "**");
  return 0;
}

/* Dynamic set the transmit gain (in dB). The origin and range are
   driver dependent.

   Note: this function is only used for user supplied dynamic
   adjustements.
*/
void trx_set_tx_gain_func(TRXState *s, double gain, int channel_num) {
  log_info("DEBUG", "trx_set_tx_gain_func");
}

/* Dynamic set the receive gain (in dB). The origin and range are
   driver dependent.

   Note: this function is only used for user supplied dynamic
   adjustements.
*/
void trx_set_rx_gain_func(TRXState *s, double gain, int channel_num) {
  log_info("DEBUG", "trx_set_rx_gain_func");
}

/* Return the maximum number of samples per TX packet. Called by
 * the application after trx_start_func.
 * Optional
 */
int trx_get_tx_samples_per_packet_func(TRXState *s) {
  log_info("DEBUG", "trx_get_tx_samples_per_packet_func");
  return 0;
}

/* Return some statistics. Return 0 if OK, < 0 if not available. */
int trx_get_stats(TRXState *s, TRXStatistics *m) {
  log_info("DEBUG", "trx_get_stats");
  return 0;
}

/* Callback must allocate info buffer that will be displayed */
void trx_dump_info(TRXState *s, trx_printf_cb cb, void *opaque) {
  log_info("DEBUG", "*opaque");
}

/* Return the absolute TX power in dBm for the TX channel
   'channel_num' assuming a square signal of maximum
   amplitude. This function can be called from any thread and
   needs to be fast. Return 0 if OK, -1 if the result is not
   available. */
int trx_get_abs_tx_power_func(TRXState *s,
                                 float *presult, int channel_num) {
  log_info("DEBUG", "trx_get_abs_tx_power_func");
  return 0;
}

/* Return the absolute RX power in dBm for the RX channel
   'channel_num' assuming a square signal of maximum
   amplitude. This function can be called from any thread and
   needs to be fast. Return 0 if OK, -1 if the result is not
   available. */
int trx_get_abs_rx_power_func(TRXState *s,
                                 float *presult, int channel_num) {
  log_info("DEBUG", "trx_get_abs_rx_power_func");
  return 0;
}

/* Remote API communication
 * Available since API v14
 * trx_msg_recv_func: called for each trx received messages
 * trx_msg_send_func: call it to send trx messages (They must be registered by client)
 * For each message, a call to send API must be done
 */
void trx_msg_recv_func(TRXState *s, TRXMsg *msg) {
  log_info("DEBUG", "trx_msg_recv_func");
}

TRXMsg* trx_msg_send_func(TRXState *s) {
  log_info("DEBUG", "trx_msg_send_func");
  return NULL;
}

/* Return actual transmit gain (in dB). The origin and range are
   driver dependent.
*/
void trx_get_tx_gain_func(TRXState *s, double *gain, int channel_num) {
  log_info("DEBUG", "trx_get_tx_gain_func");
}

/* Returns actual receive gain (in dB). The origin and range are
   driver dependent.
*/
void trx_get_rx_gain_func(TRXState *s, double *gain, int channel_num) {
  log_info("DEBUG", "trx_get_rx_gain_func");
}

/* Stop operation of the transceiver - to be called after trx_start.
   resources allocated in init are not released, so trx_call can be called again */
void trx_stop_func(TRXState *s) {
  log_info("DEBUG", "trx_stop_func");
}

/* OFDM mode: experimental 7.2 API */

/* read the current timestamp (only used in OFDM mode). Return 0
   if OK, < 0 if not supported by device. */
int trx_read_timestamp(TRXState *s, trx_timestamp_t *ptimestamp,
                          int port_index) {
  log_info("DEBUG", "trx_read_timestamp");
  return 0;
}

void trx_set_tx_streams(TRXState *s, int rf_port_index,
                           int n_streams, const TRXOFDMStreamInfo *streams) {
  log_info("DEBUG", "trx_set_tx_streams");
}
void trx_set_rx_streams(TRXState *s, int rf_port_index,
                           int n_streams, const TRXOFDMStreamInfo *streams) {
  log_info("DEBUG", "trx_set_rx_streams");
}
/* schedule the reading of OFDM symbols. Return 0 if OK, < 0 if
   error. */
int trx_schedule_read(TRXState *s, int rf_port_index,
                         const TRXScheduledSymbol *symbols,
                         int n_symbols) {
  log_info("DEBUG", "trx_schedule_read");
  return 0;
}

/* AGC functions */
int trx_set_agc_func(TRXState *s, const TRXAGCParams *p, int channel) {
  log_info("DEBUG", "trx_set_agc_func");
  return 0;
}
int trx_get_agc_func(TRXState *s, TRXAGCParams *p, int channel) {
  log_info("DEBUG", "trx_get_agc_func");
  return 0;
}

int trx_driver_init(TRXState *s1)
{
  TRXEcpriState *s;
  double val;

  // Lock all current and future pages from preventing of being paged to swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    log_error("TRX_ECPRI", "mlockall failed");
  }

  log_info("TRX_ECPRI", "Init");

  if (s1->trx_api_version != TRX_API_VERSION) {
    fprintf(stderr, "ABI compatibility mismatch between LTEENB and TRX driver (LTEENB ABI version=%d, TRX driver ABI version=%d)\n",
        s1->trx_api_version, TRX_API_VERSION);
    return -1;
  }

  s = malloc(sizeof(TRXEcpriState));
  memset(s, 0, sizeof(*s));

  trx_get_param_double(s1, &val, "recv_affinity");
  s->recv_affinity = (int) val;
  trx_get_param_double(s1, &val, "send_affinity");
  s->send_affinity = (int) val;
  trx_get_param_double(s1, &val, "encode_affinity");
  s->encode_affinity = (int) val;
  trx_get_param_double(s1, &val, "decode_affinity");
  s->decode_affinity = (int) val;
  trx_get_param_double(s1, &val, "statistic_affinity");
  s->statistic_affinity = (int) val;
  trx_get_param_double(s1, &val, "flow_id");
  s->flow_id = (int) val;
  trx_get_param_double(s1, &val, "frame_frequency");
  s->frame_frequency = (int) val;
  trx_get_param_double(s1, &val, "trx_buf_size");
  s->trx_buf_size = (int) val;
  trx_get_param_double(s1, &val, "txrx_buf_size");
  s->txrx_buf_size = (int) val;
  trx_get_param_double(s1, &val, "trace_rx");
  s->trace_rx = (int) val;
  trx_get_param_double(s1, &val, "trace_tx");
  s->trace_tx = (int) val;
  trx_get_param_double(s1, &val, "trace_offset");
  s->trace_offset = (int) val;
  trx_get_param_double(s1, &val, "monitor_pps");
  s->monitor_pps = (int) val;
  trx_get_param_double(s1, &val, "monitor_trigger_duration");
  s->monitor_trigger_duration = (int) val;
  trx_get_param_double(s1, &val, "start_sending");
  s->start_sending = (int) val;
  trx_get_param_double(s1, &val, "start_receiving");
  s->start_receiving = (int) val;
  trx_get_param_double(s1, &val, "rx_n_channel");
  s->rx_n_channel = (int) val;
  trx_get_param_double(s1, &val, "tx_n_channel");
  s->tx_n_channel = (int) val;
  trx_get_param_double(s1, &val, "statistics_refresh_rate_ns");
  s->statistics_refresh_rate_ns = (int) val;
  trx_get_param_double(s1, &val, "ecpri_period");
  s->ecpri_period = (int) val;

  if(s->ecpri_period == 0)
    log_exit("TRX_ECPRI", "ecpri_period parameter can't be null\n");
  if(s->rx_n_channel == 0)
    log_exit("TRX_ECPRI", "rx_n_channel parameter can't be null\n");
  if(s->tx_n_channel == 0)
    log_exit("TRX_ECPRI", "tx_n_channel parameter can't be null\n");

  s->rrh_mac = (uint8_t *) trx_get_param_string(s1, "rrh_mac");
  s->bbu_mac = (uint8_t *) trx_get_param_string(s1, "bbu_mac");
  s->bbu_if = (uint8_t *) trx_get_param_string(s1, "bbu_if");
  s->dpdk_options = trx_get_param_string(s1, "dpdk_options");
  s->log_directory = (uint8_t *) trx_get_param_string(s1, "log_directory");

  recv_pps_threshold = (s->frame_frequency * 9 / 10);

  s1->opaque = s;
  s1->trx_end_func = trx_ecpri_end;
  s1->trx_write_func2 = trx_ecpri_write;
  s1->trx_read_func2 = trx_ecpri_read;
  s1->trx_start_func = trx_ecpri_start;
  s1->trx_get_sample_rate_func = trx_ecpri_get_sample_rate;

  //s1->trx_write_func = trx_write_func;
  //s1->trx_read_func = trx_read_func;
  //s1->trx_set_tx_gain_func = trx_set_tx_gain_func;
  //s1->trx_set_rx_gain_func = trx_set_rx_gain_func;
  //s1->trx_get_tx_samples_per_packet_func = trx_get_tx_samples_per_packet_func;
  //s1->trx_get_stats = trx_get_stats;
  //s1->trx_dump_info = trx_dump_info;
  //s1->trx_get_abs_tx_power_func = trx_get_abs_tx_power_func;
  //s1->trx_get_abs_rx_power_func = trx_get_abs_rx_power_func;
  //s1->trx_msg_recv_func = trx_msg_recv_func;
  //s1->trx_get_tx_gain_func = trx_get_tx_gain_func;
  //s1->trx_get_rx_gain_func = trx_get_rx_gain_func;
  //s1->trx_stop_func = trx_stop_func;
  //s1->trx_read_timestamp = trx_read_timestamp;
  //s1->trx_set_tx_streams = trx_set_tx_streams;
  //s1->trx_set_rx_streams = trx_set_rx_streams;
  //s1->trx_schedule_read = trx_schedule_read;
  //s1->trx_set_agc_func = trx_set_agc_func;
  //s1->trx_get_agc_func = trx_get_agc_func;
  //s1->trx_msg_send_func = trx_msg_send_func;

  return 0;
}
